# -*- coding:utf-8 -*-
# Created by qinwei on 2017/11/29
#
import abc

from hunters.config import DEFAULT_BROWSER_CONFIG


class Factory:
    """实例工厂, 负责创建实例, 并且也负责销毁, 谁创建谁知道怎么销毁"""

    def __init__(self, browser_config=DEFAULT_BROWSER_CONFIG):
        self._config = browser_config

    @abc.abstractmethod
    def new_instance(self):
        """
        创建新的实例
        """

    @abc.abstractmethod
    def destroy(self, obj):
        """
        Factory让你决定如何创建和如何销毁,
        :param obj:
        :return:
        """

    @abc.abstractmethod
    def destroy_all(self, pool_list=None):
        """
         完全销毁所有实例
        :param pool_list: all resource pool data which is the `new_instance` method create
        :return:
        """

    def config(self):
        """
        return the config for use create instance if has
        :return:
        """
        return self._config



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/8
import unittest

from hunters.browser import ChromeTab, BaseChrome
from hunters.chrome import ChromeTabDevToolFactory
from hunters.config import BrowserConfig


class TestBrowser(unittest.TestCase):
    def setUp(self):
        canary = r"C:\Users\ADMIN\AppData\Local\Google\Chrome SxS\Application\chrome.exe"
        chrome = r"C:\Program Files (x86)\Google\Chrome\Application\chrome.exe"
        self.config = BrowserConfig(
            execute_bin=chrome,
            headless=False)

    def test_factory(self):
        factory = ChromeTabDevToolFactory(browser_config=self.config)
        chrome = BaseChrome(factory=factory)
        for i in range(2):
            r = chrome.get("https://news.qq.com/a/20171124/000144.htm")
            # r = chrome.get("https://baidu.com")
            r.window_size(1000, 1000)
            r.scroll_by(0, 10000)
            print(r.execute_js("document.title"))
            print(r.text)
            print(r.cookies(urls=["baidu.com"]))
            print(r.get_all_cookies())
            print(i)

    def test_multi_chrometab(self):
        cc = ChromeTab(browser_config=self.config)
        cc2 = ChromeTab(browser_config=self.config)
        r = cc.get("http://baidu.com")
        r2 = cc2.get("http://qq.com")
        print(r.text)


if __name__ == '__main__':
    TestBrowser.main()



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/7

##############################################################################
import base64
import logging
import os
import queue
import socket
import threading
import weakref
from base64 import b64decode
from http.cookiejar import split_header_words
from parser import ParserError
from time import sleep, time
from urllib import parse

import pychrome
import requests
from requests import Response
from requests.exceptions import TooManyRedirects, SSLError
from requests.structures import CaseInsensitiveDict

from hunters.benchmark import BenchMark
from hunters.config import DEFAULT_BROWSER_CONFIG
from hunters.constant import Regex
from hunters.util.file_util import FileUtils
from hunters.utils import StringUtil, ResponseUtils, remove_xml_encoding, dom

logger = logging.getLogger("browser")


class Browser(object):
    """ Define Browser uniform Abstract API """

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.quit()

    def get(self, url, **kwargs):
        """

        :param url:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def post(self, url, data, json, **kwargs):
        """

        :param url:
        :param data:
        :param json:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    def head(self, url, **kwargs):
        raise NotImplementedError

    def screenshot_as_png(self, filename, full_page=False):
        """

        :param filename:
        :param full_page:
        :return:
        """
        raise NotImplementedError

    def window_size(self, width, height):
        """
        resize window size

        :param width:
        :param height:
        :return:
        """
        raise NotImplementedError

    def scroll_by(self, x, y):
        """

        :param x:
        :param y:
        :return:
        """
        raise NotImplementedError

    def wait(self, num):
        """

        :param num:
        :return:
        """
        raise NotImplementedError

    def execute_js(self, js_str):
        """

        :param js_str:
        :return:
        """
        raise NotImplementedError

    def cookies(self, urls):
        """
        Cookie相关操作接口

        :param urls: 传递一个host或者url列表来获取对应的cookie, 具体由子类实现
        :return: 返回cookies (dict或list)
        """
        raise NotImplementedError

    def close(self):
        """ 关闭当前窗口, 通常如果是Selenium实现的, 可能关闭的是浏览器 """
        raise NotImplementedError

    def quit(self):
        """ 完全关闭进程, 通常如果是Selenium, 关闭浏览器和webdriver """
        raise NotImplementedError

    def browser_config(self):
        raise NotImplementedError

    def set_block_urls(self, urls):
        """
        block url to load

        :param urls: list, ["*.png", ".jpeg"]
        :return:
        """
        raise NotImplementedError

    def limit_page_size(self, per_size, total_size):
        """
        limit page size

        :param per_size:
        :param total_size:
        :return:
        """
        raise NotImplementedError

    def disable_image(self, bool_):
        """

        :param bool_:
        :return:
        """
        raise NotImplementedError

    @staticmethod
    def parse_url(url):
        """
        ParseResult(scheme='https', netloc='fex.bdstatic.com', path='/hunter/alog/dp.csp.min.js', params='', query='v=140804', fragment='')

        :param url:
        :return:
        """
        return parse.urlparse(url)

    @staticmethod
    def get_host(url):
        return parse.urlparse(url).netloc


class MiniBrowser(Browser):
    """
    文本型/单请求/微浏览器实现,
    基于requests,
    requests, Cookie自动保持, URL跳转自动跟随, 自动解析文本decode, 都是浏览器必有特性
    decode比较弱, 自己重新实现了一个, 更能准确判断gbk和utf8
    """

    def __init__(self, browser_config=DEFAULT_BROWSER_CONFIG):
        # self.session.mount("data:", DataAdapter) TOTO
        self._browser_config = browser_config

    @property
    def session(self):
        # XXX 解决同一个session长期复用会导致内存一致涨的问题
        session = requests.session()
        session.headers['User-Agent'] = self._browser_config.user_agent
        session.max_redirects = self._browser_config.max_redirects
        return session

    @property
    def cookie_store(self):
        return self.browser_config().cookie_store

    def browser_config(self):
        return self._browser_config

    def head(self, url, **kwargs):
        try:
            kwargs.setdefault("timeout", 2)  # head 请求如果2秒没啥返回, 说明拒绝head或者访问失败了
            assert url.startswith("http"), "HEAD only support http or https"
            return self.session.head(url, **kwargs)
        except Exception as e:
            logger.error("url[%s], msg[%s]" % (url, e))
            r = Response()
            r.headers = {}
            r.url = url
            r.status_code = 503  #: 服务无法访问
            return r

    def cookies_str(self, host):
        cookies_dict = self.cookies(host)
        result = ""

        for (k, v) in cookies_dict.items():
            result += "%s=%s; " % (StringUtil.replace_blank(k), StringUtil.replace_blank(v))

        return result

    def cookies(self, host):
        """从会话中拿出cookie"""
        return self.cookie_store.get(host) or {}

    def _mrg_cookies(self, host, cookies):
        """
        合并cookies
        :param cookies: list[tuple] or dict or "A=1; B=3"
        :return:
        """
        last_cookies = self.cookies(host)
        if isinstance(cookies, dict):
            last_cookies.update(cookies)
        else:
            if isinstance(cookies, str):
                #: split word ["A=1"] ==> [[(A, 1)]]
                cookies = split_header_words([cookies])[0]

            for item in cookies:
                last_cookies.update({item[0]: item[1]})

        self.cookie_store.set(host, last_cookies)
        return last_cookies

    def _set_header_cookie(self, host, **kwargs):

        headers = kwargs.get("headers", {})
        user_cookie = headers.get("Cookie")  #: 用户自己配置Cookie的情况下, 合并
        if user_cookie:
            self._mrg_cookies(host, user_cookie)
        headers.update({'Cookie': self.cookies_str(host)})
        kwargs.setdefault("headers", headers)
        return kwargs

    def _check_cookies(self, host, response, cookie_dict=None):
        if "Set-Cookie" in response.headers:
            self._mrg_cookies(host, cookie_dict)  #: 合并到会话级别, cookies中

    def get(self, url, **kwargs):
        b = BenchMark(__name__)
        try:
            # extract_cookies_to_jar(self.session.cookies, req, urllib3.HTTPResponse())
            host = self.get_host(url)
            kwargs = self._set_header_cookie(host, **kwargs)
            kwargs.setdefault("stream", True)  #: lazyLoad , 避免读取大文件
            with self.session as session:
                try:
                    r = session.get(url, **kwargs)
                except SSLError as err:  # 这里可能会报Https鉴权错误
                    kwargs.setdefault("verify", False)
                    r = session.get(url, **kwargs)
                    setattr(r, "SSLError", True)  # 配置一个标记说明存在鉴权, 让后续者鉴定错误
                    logger.warning("SSLError, url[%s], err[%s]", url, err)

                self._check_cookies(host, r, session.cookies.items())

                self._check_body_size(response=r)

            return self.__detect_encoding(r)

        except TooManyRedirects as err:
            r = ResponseUtils.too_many_redirects_response(err)
            r.ori_url = url
            logging.warning("TooManyRedirects:[{}]".format(url))
            return r
        except Exception as e:
            #: 如果是requests都无法访问, 说明连接不可用.
            r = ResponseUtils.exception_response(e)
            r.ori_url = url
            logger.error("connect url:[{}] ,error[{}]".format(url, e))
            return r
        finally:
            b.mark("OK")

    def _check_body_size(self, response):
        """ 检测内容是否超标 """
        if ResponseUtils.is_body_over_size(response=response, max_size=self.browser_config().max_body_size):
            response.close()

    def post(self, url, data=None, json=None, **kwargs):
        host = self.get_host(url)
        self._set_header_cookie(host, **kwargs)
        with self.session as session:
            r = self.session.post(url, data, json, **kwargs)
            self._check_cookies(host, r, session.cookies.items())
        return self.__detect_encoding(r)

    def screenshot_as_png(self, filename, full_page=False):
        raise NotImplementedError("IGNORE screenshot_as_png for textBrowser")

    def window_size(self, width, height):
        raise NotImplementedError("IGNORE window_size for textBrowser")

    def scroll_by(self, x, y):
        raise NotImplementedError("IGNORE scroll_by for textBrowser")

    def wait(self, num):
        raise NotImplementedError("IGNORE scroll_by for textBrowser")

    def execute_js(self, js_str):
        raise NotImplementedError("IGNORE execute_js for textBrowser")

    @staticmethod
    def __detect_encoding(response):
        """ 尝试检测页面编码 """
        #: 'text/html; charset=GB2312'
        #: 'application/json'
        content_type = response.headers.get("Content-Type", "error/no-content-type")
        if not Regex.RE_TYPE_PLAIN.search(content_type):
            #: 必须是文本类型才能检测, 二进制的忽略
            return response

        before_encoding = response.encoding  #: 'ISO-8859-1'
        detect_encoding = None

        #: 有的网站不按常理出牌, 返回content-type是charset=gbk,但是内容meta是utf8,比如QQ某个网站
        #: 但是调试发现浏览器有优先级, 如果返回头返回了编码, 以返回头为准, 否则看页面的meta[charset]值
        match = Regex.RE_CONTENT_TYPE_CHARSET.search(content_type)
        if match:
            detect_encoding = match.group(1)

        elif Regex.RE_TYPE_HTML.search(content_type):
            match = Regex.RE_MATA_CHARSET.search(response.text)
            if match:
                detect_encoding = match.group(1)  # 解决乱码问题

        #: 如果都没有编码信息, 就探测编码
        if detect_encoding is None and before_encoding in (
                None, 'ISO-8859-1'):  #: 'ISO-8859-1' 编码特别不准. 至少GBK, UTF-8也兼容他
            detect_encoding = response.apparent_encoding  #: 检测编码

        if detect_encoding is None:
            detect_encoding = "UTF-8"
        #: GB18030 (7W字) > GBK (2W字) > BIG5(繁体) > GB2312 (6K字)
        #: 如果检测出使用GB2312 , 有超出范围的罕见字符会有乱码, 改成兼容性更强的GBK
        if detect_encoding.upper() == "GB2312":
            detect_encoding = "GBK"

        response.encoding = detect_encoding
        logging.debug("DETECT ENCODING, url[%s], encoding[%s], before[%s]", response.url, response.encoding,
                      before_encoding)
        return response


##################################### ViewBrowser ##################################################

class BaseChrome(Browser):
    """ Chrome implement , base on DevTools Protocol, use DevTools(Remote-Debugging) directly, no WebDriver  """

    def __init__(self, factory):
        """
        :param factory:  devtool factory
        """
        self._browser_config = factory.config()
        self._minibrowser = MiniBrowser(self._browser_config)
        self.local_store = self._browser_config.local_store
        self._last_done = True
        self._query_done = queue.Queue(1)
        self._factory = factory
        self._response_received_queue = queue.Queue()
        self._listen_count = False
        self.api = factory.new_instance()
        self._max_total_size = self.browser_config().max_body_size * 5
        self._init()

    def _init(self):

        self.Page.enable()
        self.DOM.enable()
        self.Network.enable()
        try:
            # This method is EXPERIMENTAL
            self.api.Page.setDownloadBehavior(behavior="allow", downloadPath=self.local_store)

            if self.browser_config().max_body_size > 0:
                self.api.Network.setDataSizeLimitsForTest(maxTotalSize=self._max_total_size,
                                                          maxResourceSize=self.browser_config().max_body_size)

            def response(**kwargs):
                if self._listen_count > 0:
                    self._response_received_queue.put(kwargs)
                    self._listen_count -= 1
                else:
                    logger.debug("ignore {}", kwargs)

            self.api.Network.responseReceived = response

        except Exception as e:
            logger.warning(str(e))

        self.Page.loadEventFired = self.__event_page_load

    def browser_config(self):
        return self._browser_config

    @property
    def text(self):
        data = self.DOM.getDocument()
        node_id = data.get("root").get("nodeId")
        result_dict = self.DOM.getOuterHTML(nodeId=node_id)
        return result_dict.get("outerHTML")

    def get(self, url, **kwargs):
        """
        Because Connect to Chrome base on WebSocket, so when we send "Page.navigate" command to browser,
        we should be listening `PageLoadedFired` Event to known whether the all page load done

        :param url:
        :param kwargs:
        :return: Response
        """
        b = BenchMark("chrome-get")
        while not self._last_done:
            # block to wait for last query done
            sleep(0.01)

        self._last_done = False

        # 某些网站卡在https验证部分都没有发请求
        _timeout = kwargs.get("timeout", 3)
        _headers = kwargs.get("headers", {})
        _referer = _headers.get("Referer")
        _cookies = _headers.get("Cookies")
        # TODO cookie headers
        response = self._minibrowser.get(url, **kwargs)
        try:
            # for open or query timeout
            self.Page.navigate(url=url, referer=_referer, _timeout=_timeout)

            # for page load timeout
            self._query_done.get(block=True, timeout=_timeout)
            response.status_code = 200

        except (queue.Empty, pychrome.exceptions.TimeoutException) as e:
            # force browser stop
            self._last_done = True
            response.status_code = -5404
            response.reason = "timeout"
            logger.info("STOP LOADING [{}]".format(url))
            self.api.Page.stopLoading()

        response.url = self.url or url  # current page
        b.ok()
        tab = Tab(weakref.ref(self)(), weakref.ref(response)())
        setattr(tab, "headless", True)
        return tab

    @property
    def url(self):
        url = self.execute_js("location.href")
        if url.startswith("chrome-error"):
            return None
        return url

    def wait(self, num):
        sleep(num)
        return self

    def post(self, url, data, json, **kwargs):
        raise NotImplemented("Not Support POST method")

    def close(self):
        return self

    def scroll_by(self, x, y):
        """ scroll the scroll bar, to trigger some async page load"""
        return self.execute_js("window.scrollBy({}, {})".format(x, y))

    def head(self, url, **kwargs):
        return self._minibrowser.head(url, **kwargs)

    def execute_js(self, js_str):
        """
        :param js_str: js expression
        :return: the js expression return (maybe u should specify "return document.title"
        """
        b = BenchMark("execute_js")
        try:
            result = self.api.Runtime.evaluate(expression=js_str)
            return result.get("result").get("value", None)
        finally:
            b.ok()

    def screenshot_as_png(self, filename=None, full_page=False):
        """
        screen_shot current page

        :param filename: filename (string) to  write to
        :param full_page:
        :return: return bytes when no filename specify
        """
        b = BenchMark("screenshot_as_png")
        if full_page:
            self._resize_to_model_height("html")
        try:
            result = self.api.Page.captureScreenshot(format="png", fromSurface=True, fullPage=True)
            base64 = result.get('data', "")
            image_bytes_ = b64decode(base64)

            if filename:
                filename = os.path.join(self._browser_config.local_store, filename)
                if filename.startswith(self._browser_config.local_store) and "../" not in filename:
                    with open(filename, "wb+") as f:
                        f.write(image_bytes_)
                else:
                    raise PermissionError("deny to write outside %s" % self.local_store)

            return image_bytes_
        finally:
            self.api.Emulation.clearDeviceMetricsOverride()  # reset window size
            b.mark("ok")

    def _resize_to_model_height(self, selector):
        """

        :param selector: CSS Selector
        :return:
        """
        data = self.api.DOM.getDocument()
        node_id = data.get("root").get("nodeId")
        result = self.api.DOM.querySelector(selector=selector, nodeId=node_id)
        # result = {'model': {'content': [0, 0, 1280, 0, 1280, 5527, 0, 5527],
        # 'padding': [0, 0, 1280, 0, 1280, 5527, 0, 5527],
        # 'border': [0, 0, 1280, 0, 1280, 5527, 0, 5527],
        # 'margin': [0, 0, 1280, 0, 1280, 5527, 0, 5527],
        # 'width': 1280, 'height': 5527}}
        if "nodeId" in result:
            result = self.api.DOM.getBoxModel(nodeId=result.get("nodeId"))
            model = result.get("model")
            self.window_size(self.browser_config().width, model.get("height") + 100)

    def window_size(self, width, height):
        """ change current page size by Chrome Emulation """
        self.api.Emulation.setDeviceMetricsOverride(width=width, height=height, deviceScaleFactor=1,
                                                    screenWidth=width, screenHeight=height,
                                                    mobile=False, fitWindow=False)
        self.api.Emulation.setVisibleSize(width=width, height=height)
        self.api.Emulation.setPageScaleFactor(pageScaleFactor=1)
        return self

    def disable_image(self, bool_):
        """

        :NOTE: If use  Tab pool return a tab as a resource,
        remember to reset all behavior to protect Resource Pollution when user change the tab behavior.
        Because the pool return a blocked tab in the next time, maybe not expected
        """
        if bool_:
            # block by some url suffix
            self.set_block_urls(urls=[
                ".jpg", ".png", ".ico", ".gif", ".jpeg", "img="
                , "=png", "=jpg", "=jpeg", "=gif"])
        else:
            self.set_block_urls(urls=[])

    def limit_page_size(self, per_size, total_size):
        """
        Be careful to reset all behavior if necessary

        :param per_size: bytes
        :param total_size: bytes
        :return:
        """
        if per_size > 0 and total_size > 0:
            self.api.Network.setDataSizeLimitsForTest(maxTotalSize=total_size, maxResourceSize=per_size)

    def set_block_urls(self, urls):
        self.api.Network.setBlockedURLs(urls=urls)
        return self

    def quit(self):
        """ Close Chrome , Stop WebSocket """
        self.api.Browser.close()
        self.api.stop()

    def cookies(self, urls):
        """
        :param urls: url list , or domain list
        :return: cookies [{name: "", value:"", url:"", domain:"" ...}]
        """
        result = self.Network.getCookies(urls=urls)
        return result.get("cookies")

    def get_all_cookies(self):
        """ return a list for all cookie """
        result = self.Network.getAllCookies()
        return result.get("cookies")

    def delete_cookies(self, name, **kwargs):
        """
        :param name: cookie name
        :param kwargs:  (like string url, domain, path)
        :return:
        """
        kwargs.setdefault("name", name)
        self.api.Network.delete_cookies(**kwargs)
        return self

    def set_cookies(self, cookies):
        """
        :param cookies: [{name:"key", value:"value", url:"option", "domain": "option" }]
        :return:
        """
        self.api.Network.setCookies(cookies=cookies)
        return self

    def __event_page_load(self, **kwargs):
        """
        Page Loader Event Listener,
        emit when Chrome Page Load (all ajax, and DOM change done)

        :param kwargs:
        :return:
        """
        self._last_done = True
        self._query_done.put(time())

    def __getattr__(self, item):
        """
        Proxy Chrome DevTools Method , When Attribute is uppercase start ,
        such as self.Page.enable() bind to self.api
        """
        if item[0].isupper():
            return getattr(self.api, item)

        raise AttributeError(item)

    def listen_response(self, max_count=100):
        self._listen_count = max_count

    def list_response(self):
        """

        :return:
        """
        result = []
        # not use yield and iter because we should remove all response in
        try:
            # get all item in queue, to clear it
            while True:
                # just wait a little time
                # chrome will load new Network by js or setTimeout/setInterval
                # discard it
                item = self._response_received_queue.get(timeout=0.5)
                result.append(ChromeResponse(weakref.ref(self)(), item))

        except queue.Empty as e:
            pass

        return result


class ChromeResponse(Response):
    def __init__(self, chrome, received):
        super().__init__()
        self.request_id = received.get("requestId")
        self._response = received.get("response")
        self.headers = CaseInsensitiveDict(data=self._response.get("headers"))
        self._type = received.get("type")
        self.url = self._response.get("url")
        self.status_code = self._response.get("status")
        self.reason = self._response.get("statusText")
        self._browser = chrome

    @property
    def type(self):
        return self._type.lower()

    @property
    def text(self):
        #  {"body":"",  "base64Encoded": True/False }
        return self.content.decode()

    @property
    def content(self):
        try:
            result = self._browser.Network.getResponseBody(requestId=self.request_id)
            if result.get("base64Encoded"):
                return base64.b64decode(result.get("body"))
            return bytes(result.get("body"), encoding="utf-8")
        except Exception as e:
            print(e)
            return None


class Chrome(BaseChrome):
    """
    Multi Chrome Process
    Every Chrome Instance is create a new Chrome Process and return a Tab
    """
    factory = None

    def __init__(self, browser_config):
        if Chrome.factory is None:
            from hunters.chrome import ChromeDevToolFactory
            Chrome.factory = ChromeDevToolFactory(browser_config=browser_config)

        super().__init__(factory=Chrome.factory)

    @staticmethod
    def find_free_port():
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(('', 0))
            return s.getsockname()[1]


class ChromeTab(BaseChrome):
    """
    One Chrome Process, open Multi Tab
    """

    # Share the factory for all ChromeTab Instance,
    # In order to use a single Chrome Browser,
    # In this case, every ChromeTab instance is bind to a Chrome Tab
    factory = None

    def __init__(self, browser_config):
        if ChromeTab.factory is None:
            from hunters.chrome import ChromeTabDevToolFactory
            ChromeTab.factory = ChromeTabDevToolFactory(browser_config=browser_config)
        else:
            logger.warning("ChromeTab.factory has already set before. the Tab type will only use one Chrome Instance")

        super().__init__(factory=ChromeTab.factory)


class ViewBrowser(ChromeTab):
    """The View Browser is extend by ChromeTab
    Every New ViewBrowser Instance will create a new Tab in backend Chrome
    The browserConfig will share in all Tab
    """

    def __init__(self, browser_config):
        super().__init__(browser_config=browser_config)


###########################################################
class Tab(Browser):
    """
    Tab是网页标签, 比较特殊的东西, 不知道取什么名字, 就叫tab(标签)类似浏览器标签
    有浏览器的特性(可以继续发起请求), 也有结果的特性(从页面获取值)
    Proxy Browser/Response Method
    Friendly TO IDE code autocomplete,
    some method not use __getattr__
    """

    def __init__(self, browser, response):
        self._browser = browser
        self._response = response
        self.local_store = browser.browser_config().local_store
        self._thread_local = threading.local()
        self.__dict__.update(response.__dict__)  #: 拷贝结果集的属性到tab
        self._dom = None

    def dom(self):
        # 解析结构化结构, DOM(方便CSS方式选择)
        # 这里会牺牲一定性能来保证API的简洁
        if self._dom is None and Regex.RE_TYPE_HTML.search(self.headers.get("Content-Type")):
            text = self.text or ""
            if "" == text.strip():
                text = "<html></html>"
            try:
                self._dom = dom(remove_xml_encoding(text))  # cssselect
            except ParserError as e:
                logger.error("DOMError, url[%s], msg[%s], content:[%s]", self.url, e, text)
                self._dom = dom("<html></html>")

        if self._dom is None:
            logger.warning("DOM NONE! %s %s", self.url, self.text)
            self._dom = dom("<html></html>")
        return self._dom

    @property
    def text(self):
        if hasattr(self, "headless"):
            return self._browser.text
        return self._response.text

    @property
    def title(self):
        if "html" in self.headers.get("content-type", ""):
            #: 只有当前访问页面是HTML的类型才尝试获取 title
            els = self.dom().cssselect('title')
            return els[0].text if els and len(els) > 0 else ""
        return ""

    def user_dir(self, path):
        return os.path.join(self.local_store, path)

    def tmp_file(self, prefix="", suffix="", dirname=None, new=True):
        """
        create a new tmp file in user_dir
        such as in browser local store


        :param mode: open mode "w+b"
        :param prefix: tmp file prefix name
        :param suffix: tmp file suffix , such as ".png"
        :param dirname:
        :param new: [True] create a new temp file in  every call, otherwise bind to thread local
                    if False , it will return a file bind to current thread, different thread will hold
                    the different file. To avoiding multi thread write a same file
        :return: A temp open file

        :NOTE: CLOSE file will not Remove the temp File
        """
        if new:
            tmp_name = self.__temp_file(prefix=prefix, suffix=suffix, dirname=dirname)
            return self.user_dir(tmp_name)

        if not hasattr(self._thread_local, "tmp_file"):
            tmp_name = self.__temp_file(prefix=prefix, suffix=suffix, dirname=dirname)
            self._thread_local["tmp_file"] = self.user_dir(tmp_name)

        return self._thread_local.get("tmp_file")

    @staticmethod
    def __temp_file(prefix="", suffix="", dirname=""):
        tmp_name = FileUtils.create_tmp_file_name(prefix=prefix, suffix=suffix)
        if dirname is not None:
            tmp_name = "{}/{}".format(dirname, tmp_name)
        return tmp_name

    def scroll_by(self, x, y):
        self._browser.scroll_by(x, y)

    def execute_js(self, js_str):
        self._browser.execute_js(js_str)

    def get(self, url, **kwargs):
        self._browser.get(url, **kwargs)

    def window_size(self, width, height):
        self._browser.window_size(width, height)

    def close(self):
        self._browser.close()

    def screenshot_as_png(self, filename, full_page=False):
        self._browser.screenshot_as_png(filename, full_page)

    def quit(self):
        self._browser.quit()

    def wait(self, num):
        self._browser.wait(num)

    def __str__(self):
        return "<Tab[{}],[{}]>".format(self.url, self.status_code)

    def __getattr__(self, item):
        if hasattr(self._browser, item):
            return getattr(self._browser, item)
        return getattr(self._response, item)



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/9/11
import logging

logger = logging.getLogger("hunters.exception")


class OverLimitErr(Exception):
    """ Add Url Over limit """


class SpiderExceptionListener(object):
    """
    总的Exception控制器
    """

    def __init__(self):
        self._exception_handler = []
        self.init_exception()

    def init_exception(self):
        def show_exception(ex):
            logger.exception(ex)

        self.add_listener(show_exception)

    def add_listener(self, ex):
        if not callable(ex):
            raise ValueError("{} not callable")
        self._exception_handler.append(ex)

    def handle_exception(self, ex):
        for handler in self._exception_handler:
            handler(ex)


spider_exception_listener = SpiderExceptionListener()



# -*- coding:utf-8 -*-
import inspect
import logging
import queue
import re
import threading
from queue import Queue

from hunters.atomic import AtomicLong
from hunters.benchmark import BenchMark
from hunters.constant import AddUrlResultType
from hunters.exceptions import spider_exception_listener
from hunters.utils import get_real_url

logger = logging.getLogger("spider")


def dynamic_params(func, params):
    """
    动态参数解析, 根据函数声明动态的识别并返回参数集
    Python传递参数要和声明一致, 除非参数声明有默认值, 如果传递一个非声明的参数会报错.
    这里会过滤掉, 只返回函数声明的参数列表(dict)
    """
    params_desc = inspect.getfullargspec(func)
    result = {}
    for name in params_desc[0]:  #: 在函数声明
        if name in params:  #: 并且在规定注入列表中
            result[name] = params.get(name)
    return result


class TaskMeta(object):
    """
    任务模式下的描述数据
    也是Spider运行时候的上下文,
    在手动调用add_url时, 为保持上下文关系, 应该继续传递下去, 否则在递归或者循环中失去了上下文联系
    """

    __attrs__ = [
        "task_id", "count", "max_count", "max_deep"
    ]

    def __init__(self, task_id="default", max_count=10, max_deep=3, count=0):
        self.mutex = threading.Lock()
        self.max_deep = max_deep
        self.max_count = max_count
        self.count = count

        #: 由于任务是多线程并发处理, 并且有的线程又会不断扫描页面生成新的URL插入队列.直到到达max_count不能再插入
        #: 而当前任务尚有很多URL是在队列中等待处理. 只有当前taskid所有产生的URL处理完才能作为任务结束.
        #: 外部系统只能通过任务剩余URL数量remain_count来作为整个任务结束的标志.
        self.remain_count = 0

        #: 当前深度
        self.deep = 0
        #: 任务ID
        self.task_id = task_id

    def to_dict(self):
        return {'max_deep': self.max_deep, 'max_count': self.max_count,
                'count': self.count, 'remain_count': self.remain_count,
                'deep': self.deep, 'task_id': self.task_id}

    @staticmethod
    def from_dict(dictdata):
        obj = {}
        for item in TaskMeta.__attrs__:
            obj[item] = dictdata.get(item)
        tm = TaskMeta(**obj)
        tm.remain_count = dictdata.get('remain_count', 0)
        tm.deep = dictdata.get("deep", 0)
        return tm

    def incr_url_count(self):
        with self.mutex:
            self.count += 1
            self.remain_count += 1

    def overlimit(self):
        #:  URL两个约束条件, URL数量上限, URL深度,
        return self.count >= self.max_count or self.deep > self.max_deep

    def extend_default_value(self, max_url, max_deep):
        """ 配置默认值, 如果task_meta原来没有配置, 就用给定的默认值 """
        self.max_count = self.max_count or max_url
        self.max_deep = self.max_deep or max_deep

    def finish_one(self):
        """ 标记完成一个 """
        with self.mutex:
            self.remain_count -= 1

    def is_lastone(self):
        """
        判断是否是最后一个
        这里还是封装一下好, 一般在output里面调用, 方便在output中对任务结束时做进一步处理, 不容易造成误解,
        到底是remain_count == 1, 还是==0 , 用户不需要关心逻辑
        除非他知道处理完output以后, remain_count才会 -1.
        """
        return self.remain_count == 1


DEFAULT_QUEUE = Queue()


class Spider(object):
    """
    基本爬虫核心
    爬虫主要实现主流程
    AddURL -> Filter(N个)-> QUEUE  -> getPage -> Output(N个) 这样的主过程
    """

    def __init__(self, browser_factory, queue=DEFAULT_QUEUE):
        """
        :param browser_factory:
        :param queue:
        """
        self._url_count = AtomicLong()  # 当前URL数量
        self._default_task_meta = TaskMeta(max_count=10, max_deep=3)
        self._threadLocal = threading.local()
        self._browser_config = browser_factory.config()
        self._url_queue = queue  # URL队列
        self._all_filters = []  # 过滤器
        self._all_outputs = []  # 输出控制器
        self._threadPool = None
        self._wait_timeout = 10
        self._request_timeout = 20  #: 20s 请求超时
        self._browser_factory = browser_factory

        self._init_browser_config()
        # 这里要区分哪些是浏览器的配置, 哪些是爬虫的配置.  最大URL, 深度, 等等是属于爬虫的逻辑.

    def _init_browser_config(self):
        self._browser_config.max_body_size = 500 * 1024  #: 最大返回大小500k

    @property
    def queue(self):
        return self._url_queue

    def request_timeout(self, num):
        self._request_timeout = num
        return self

    def wait_timeout(self, num):
        self._wait_timeout = num
        return self

    def max_deep(self, deep):
        if isinstance(deep, int):
            self._default_task_meta.max_deep = deep
        return self

    def max_urls(self, num):
        if isinstance(num, int):
            self._default_task_meta.max_count = num
        return self

    def browser(self):
        if not hasattr(self._threadLocal, "browser"):
            self._threadLocal.browser = self._browser_factory.new_instance()
        return self._threadLocal.browser

    def add_url(self, url, base_url=None, deep=0, headers=None, task_meta=None, extra=None):
        """
        添加URL到抓取队列, 如果超过限制,不会添加

        :param url: 当前页面的解析出来的URL, 可能是相对路径, ../about , /a/m/info.html
        :param base_url: 当前页面的URL, (url)都是相对于base_url, 会自动转成绝对URL
        :param deep: 当前深度
        :param task_meta 用任务id区分url爬取任务. 可以实现对整体服务的重用, 比如每个task有独立的url(计数器, 深度),
                         跟总的计数器分开
        :param headers 当前url请求指定的header, 比如UA, 会用UA进行请求.
        :param extra: 额外参数
        :return: 0 达到Limit值,不能能添加
                 1 成功添加
                 -1 过滤器过滤失败.
                 -2 参数错误
        """
        absolute_url = get_real_url(base_url, url)

        #: 如果没有任务描述,使用默认的, 后续的output需要注意将任务描述一起传递, 维持任务状态
        task_meta = task_meta or self._default_task_meta
        task_meta.deep = deep

        # URL两个约束条件, URL数量上限, URL深度, 如果URL长度为空也取消
        if len(absolute_url) == 0:
            """空白URL"""
            return AddUrlResultType.URL_ERROR

        if task_meta.overlimit():
            warn_msg = "add url[%s] overlimit, ignore taskInfo: %s" % (absolute_url, task_meta.to_dict())
            logger.info(warn_msg)
            return AddUrlResultType.OVERLIMIT

        if headers is None:
            headers = {}

        if base_url is not None:
            headers.setdefault("Referer", base_url)

        url_params = self._param_to_dict(url=absolute_url, deep=deep, headers=headers, task_meta=task_meta,
                                         extra=extra)

        filter_result = self.do_filter(**url_params)

        if AddUrlResultType.OK != filter_result:
            return filter_result

        task_meta.incr_url_count()  #: 通过了过滤器, 对应任务的URL计数器递增1

        logger.debug("Add To Queue [%s], origin:[%s], base_url:[%s] count[%s]", absolute_url, url, base_url,
                     task_meta.count)

        self.each_url_handle(**url_params)

        return AddUrlResultType.OK

    @staticmethod
    def _param_to_dict(url, deep, headers, task_meta, extra):
        #: 这些参数可以在filter/output声明的时候用依赖注入方式注入, 只需要声明参数名称相同
        return {
            'url': url,
            'deep': deep,
            'headers': headers,
            'task_meta': task_meta,
            'extra': extra or dict()
        }

    def do_filter(self, url, deep, headers, task_meta, extra):
        bm = BenchMark("all-filter")

        params_ = self._param_to_dict(url, deep, headers, task_meta, extra)

        for filter_ in self._all_filters:
            if not filter_.match(deep):
                # 如果不是匹配的深度就跳过, 每个URL都根据配置进行过滤
                continue

            dy_param = dynamic_params(filter_.real_func, params_)

            if not filter_.invoke(**dy_param):
                logging.debug("Not Pass Filter=[{}], [{}]".format(filter_.real_func.__name__, url))
                return AddUrlResultType.FILTER_UN_PASS
            bm.mark("OK")

        return AddUrlResultType.OK

    def each_url_handle(self, url, deep, headers, task_meta, extra):
        url_params = self._param_to_dict(url, deep, headers, task_meta, extra)
        self.queue.put(url_params)

    def filter(self, start_deep=1):

        """ @app.filter定义过滤器, 只有过滤器返回true才会继续执行

        @:param start_deep:  支持配置起始深度过滤, 因为第一个URL是deep=0, 有可能也需要过滤, 可以对起始URL进行过滤
        """

        def decorator(func):
            self.add_filter(func, start_deep)
            return func

        return decorator

    def get_page(self, url, headers=None):
        """ 根据URL 获取页面结果 """
        browser = self.browser()
        response = browser.get(url, headers=headers, timeout=self._request_timeout)
        response.ori_url = url  #: 记录一下原始URL, 因为有的URL重定向了. 当前页面获取的URL就是重定向后的
        return response  #: 返回一个标签页

    def output(self, content_type=r".*"):
        """ @app.output装饰器来添加输出控制器, 利用content_type指定只处理的类型 """

        def decorator(func):
            self.add_output(content_type, output=func)
            return func

        return decorator

    @staticmethod
    def _default_headers(headers):
        default_headers = {}
        default_headers.update(headers or {})
        return default_headers

    def _do_url_task_from_queue(self):

        if self._wait_timeout:
            item = self.queue.get(timeout=self._wait_timeout)
        else:
            item = self.queue.get()

        url_ = item.get('url')
        deep_ = item.get('deep', 0)
        task_meta = item.get("task_meta")  #: 必须存在
        extra_ = item.get('extra', None)  #: 额外传递的参数, 附属内容传递给filter或者output
        headers_ = item.get('headers', {})

        self.start_query(url_, deep_, headers_, task_meta, extra_)

    def start_query(self, url, deep, headers=None, task_meta=None, extra=None):
        """ 执行抓取动作, 返回结果经过output捕获 """
        bm = BenchMark("queue-item")

        default_headers = self._default_headers(headers)

        try:
            tab = self.do_url(url, deep, default_headers, task_meta, extra)

            self.do_output(tab, url=url, deep=deep, headers=headers, task_meta=task_meta, extra=extra)

        except Exception as e:  # 这里有可能网络请求抛出各种异常
            #: 有的是网页无法访问
            logger.error("GET PAGE, url[%s], msg[%s]", url, e)
            spider_exception_listener.handle_exception(e)
        finally:
            #: 从队列里取出来的URL,执行完成后, 在这个任务中标记完成了一个URL, 直到remain_count为0
            task_meta.finish_one()
            bm.mark("OK")

    def _loop_run(self):
        """ 循环从队获取URL进行处理, 多线程的情况下, 每个线程执行的方法 """
        # count = 0
        while True:
            """
            在一个死循环的作用域中, 变量不会马上回垃圾回收, 最好能封装一个方法形成局部作用域
            以便每次跳出循环都可以回收, 否则这个循环体内的方法内存会一直涨....无法GC
            因为Python可能认为, 在这个循环体中的变量都可能还会被使用, 只要这个作用域没有结束, 就不会立即回收
            """
            self._do_url_task_from_queue()

    def do_url(self, url_, deep, headers=None, task_meta=None, extra=None):
        """处理URL, 请求获取页面, 该方法可以在output中独立递归调用 """
        logger.info("GET PAGE[%s], deep[%s]", url_, deep)
        bg = BenchMark("getpage")
        tab = self.get_page(url_, headers=headers)  #: 这里可能抛出异常, 向上抛
        bg.mark("OK")

        if not tab.ok:
            # for 404
            logger.warning("[404], %s", url_)

        return tab

    def do_output(self, tab, url, deep, headers, task_meta, extra):
        #: 这里如果找不到content-type说明网络请求, 存在错误, 无返回等
        #: 定义个默认的错误内容头, 以便外部可以通过绑定error处理器, 统一通过特定output处理
        content_type = tab.headers.get('Content-Type', "error/unknown")

        bm = BenchMark("all-output")

        url_params = {  #: 这些参数可以在filter/output声明的时候用依赖注入方式注入, 只需要声明参数名称相同
            'tab': tab,
            'url': url,
            'deep': deep,
            'headers': headers,
            'task_meta': task_meta,
            'extra': extra
        }

        for output_handler_ in self._all_outputs:
            # 遍历所有的输出控制器, 如果符合控制器约束, 就执行控制器
            if output_handler_.match(content_type):
                dy_params = dynamic_params(output_handler_.real_func, url_params)
                output_handler_.invoke(**dy_params)

        try:
            tab._response.close()
        except Exception as e:
            # when 404
            pass
        bm.mark("OK")

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def add_filter(self, func, start_deep=1):
        """
        添加一个过滤器
        :param func:
        :param start_deep: 起作用的URL深度, 第一个URL为0
        :return:
        """
        if not callable(func):
            raise TypeError("Filter[{}] must be a function" % func)

        self._all_filters.append(Filter(func, start_deep=start_deep))
        return self

    def add_output(self, content_type, output):
        """
        添加一个输出控制器
        :param content_type:  正则表达式, 匹配的content-type类型
        :param output: 输出控制器方法
        :return:
        """
        if not callable(output):
            raise TypeError("Output[{}] must be a function" % output)
        self._all_outputs.append(OutputHandler(content_type, output))
        return self

    def _mutli_thread_run(self, num, wait=True):
        # 创建多线程执行, daemon使得后台进程在主线程结束后自动销毁
        # 这个方法可以被覆盖, 子类可以实现空转不停机. 具体看SpiderServer
        # Spider一开始是编程式的爬虫, 没有提供服务形式, 完整的编程功能是, 从main函数运行完直到结束
        # 服务形式的爬虫不停机运转. 队列不断接受请求.
        #
        self._threadPool = {}

        def wrap():
            try:
                self._loop_run()
            except queue.Empty as e:
                self.browser().quit()
            except Exception as e:
                logger.error("Thread[%s] run error %s " % (threading.currentThread().name, e))
                self._threadPool.pop(threading.currentThread().name)  #: 这个线程死掉了,
                self.browser().quit()
                spider_exception_listener.handle_exception(e)

        try:
            start_num = num - len(self._threadPool.keys())
            pending_threads = []
            for i in range(start_num):
                t = threading.Thread(target=wrap, args=(), daemon=False)
                self._threadPool[t.name] = t
                pending_threads.append(t)

            for t in pending_threads:
                logger.info("MuitlThreadRun[%s]" % t.name)
                t.start()

            if wait:
                # 如果存在等待, 等待子线程处理完成, 所有的子线程配合wait_timeout, 直到没有任务超时停止
                # 如果没有配置wait_timeout, 子线程就永远不会停止
                for t in pending_threads:
                    t.join()

            logger.info("=======================ALL DONE, COUNT[%s] =====================", self._url_count.value)

        except Exception as e:
            logger.exception(e)

    def run(self, num=1, wait=True):
        with self:
            self._mutli_thread_run(num, wait)
            # self._loop_run()


###########################################################
class Filter(object):
    """ 过滤器接口对象 """

    def __init__(self, func, start_deep=1):
        """

        :param func: 方法
        :param start_deep: 从哪个URL深度开启起作用, 一般第一个URL, deep=0
        """
        self.func = func
        self.start_deep = start_deep

    @property
    def real_func(self):
        return self.func

    def match(self, deep):
        return deep >= self.start_deep

    def invoke(self, *args, **kwargs):
        if callable(self.func):
            bx = BenchMark(self.func.__name__)
            try:
                return self.func(*args, **kwargs)
            except Exception as ex:
                spider_exception_listener.handle_exception(ex)
            finally:
                bx.mark("OK")
        return False

    def __call__(self, *args, **kwargs):
        #: 直接函数调用, 一般都在调试阶段, 有什么异常就抛出什么异常
        return self.func(*args, **kwargs)


###########################################################
class OutputHandler(object):
    """ 输出控制器 """

    def __init__(self, content_type, func):
        self.content_type = content_type
        self.re_content_type = re.compile(content_type)
        self.funcs = [func]

    @property
    def real_func(self):
        return self.funcs[0]

    def invoke(self, *args, **kwargs):
        b = BenchMark(self.real_func.__name__)
        try:
            #: 任何一个outputHandler抛出异常都会影响其他output handler执行, 直接抛异常, 否则很多问题无法察觉
            return self.real_func(*args, **kwargs)
        except Exception as ex:
            spider_exception_listener.handle_exception(ex)
        finally:
            b.mark("OK")

    def match(self, content_type):
        return self.re_content_type.search(content_type)

    def __call__(self, *args, **kwargs):
        #: 直接按照函数调用, 一般都在调试阶段, 有什么异常就抛出什么异常
        return self.real_func(*args, **kwargs)



# -*- coding:utf-8 -*-
# Created by qinwei on 2017/8/8
import logging
import re

from hunters.browser import MiniBrowser
from hunters.constant import Regex
from hunters.defaults import DefaultFilter

print("window.scrollBy({}, {})".format(10, 10))

filter = DefaultFilter()
FORMAT = "%(asctime)s [%(levelname)s] [%(name)s] %(module)s[%(lineno)d]-%(message)s"

logging.basicConfig(level=logging.INFO, format=FORMAT)


def lookup(func):
    print(func("url"))
    print(func("url"))
    print(func("url"))
    print(func("url"))


